import 'dart:convert';

import 'package:http/http.dart';
import 'package:puma_core/redux/core/coreAction.dart';
import 'package:puma_network/constants.dart';
import 'package:puma_network/helpers.dart';
import 'package:puma_network/redux/network/networkState.dart';
import 'package:puma_network/redux/response/responseAction.dart';
import 'package:redux/redux.dart';

abstract class APIAction implements CoreAction {
  String get method;
  bool get isFetching;
  String get networkCompletionKey;
  Uri endpoint(bool https, String apiEndpoint);
  Future<Response> apiCall(NetworkState store);

  decodeResponse(Store store, Response data);
  setNetworkCompletionKey(String networkCompletionKey);
}

abstract class CompleterAction implements CallbackAction, ErrorCallbackAction {}

abstract class CallbackAction {
  Function()? get callback;
}

abstract class ErrorCallbackAction {
  Function(Error, [StackTrace]) get errorCallback;
}

class RequestAction implements APIAction {
  final String method;
  final bool isFetching;
  final int timestamp;
  final String dispatcherKey;
  final String authority;
  String path;
  String networkCompletionKey = "";

  Map<String, String>? queryParams;

  setNetworkCompletionKey(String networkCompletionKey) =>
      this.networkCompletionKey = networkCompletionKey;

  decodeResponse(Store store, Response data) => data;

  Map<String, String> headers(NetworkState store) {
    Map<String, String> headers = {
      "$headerTokenKey": store.token,
      "Accept": "*/*",
      "Accept-Encoding": "gzip, deflate",
      "content-type": "application/json"
    };

    if (store.authImpersonateToken != null) {
      headers["$headerImpersonateTokenKey"] = store.authImpersonateToken!;
    }

    print("headers: $headers");

    return headers;
  }

  bool operator ==(a) => a is RequestAction && hashCode == a.hashCode;
  int get hashCode =>
      timestamp.hashCode ^ method.hashCode ^ authority.hashCode ^ path.hashCode;

  Uri endpoint(bool https, String apiEndpoint) {
    if (queryParams == null) {
      return uriBuilder(https, apiEndpoint, path);
    }
    return uriBuilderParams(https, apiEndpoint, path, queryParams!);
  }

// Placeholder Method
  apiCall(NetworkState store) async =>
      await post(this.endpoint(store.https, store.apiEndpoint),
          headers: headers(store),
          encoding: Encoding.getByName("application/json"));

  RequestAction(this.path, this.method,
      {this.authority = "",
      this.dispatcherKey = "",
      this.queryParams,
      this.isFetching = false})
      : this.timestamp = DateTime.now().millisecondsSinceEpoch;
}

abstract class Responder {
  CoreAction call(model);
}

class TypedResponse<Model> implements Responder {
  final CoreAction Function(
    Model model,
  ) middleware;

  TypedResponse(this.middleware);

  CoreAction call(dynamic model) {
    if (model is Model) {
      return middleware(model);
    } else {
      return EmptyResponse();
    }
  }
}

class PostAction extends RequestAction {
  final Object body;
  final String url;

  Future<Response> apiCall(NetworkState state) async {
    Uri finalEndpoint = endpoint(state.https, getFinalEndpoint(url, state));
    print("POST $finalEndpoint");

    return await post(
      finalEndpoint,
      body: jsonEncode(body),
      headers: headers(state),
      encoding: Encoding.getByName("application/json"),
    );
  }

  PostAction(String path, this.body, {this.url = ""}) : super(path, "POST");
}

class GetAction extends RequestAction {
  final String url;

  Future<Response> apiCall(NetworkState state) async {
    Uri finalEndpoint = endpoint(state.https, getFinalEndpoint(url, state));
    print("GET $finalEndpoint");

    return await get(
      finalEndpoint,
      headers: headers(state),
    );
  }

  GetAction(String path, {this.url = ""})
      : super(path, "GET", isFetching: true);
}

class DeleteAction extends RequestAction {
  final String url;
  Future<Response> apiCall(NetworkState state) async {
    Uri finalEndpoint = endpoint(state.https, getFinalEndpoint(url, state));
    print("DELETE $finalEndpoint");
    return await delete(
      finalEndpoint,
      headers: headers(state),
    );
  }

  DeleteAction(String path, {this.url = ""}) : super(path, "DELETE");
}

class PutAction extends RequestAction {
  final Object body;
  final String url;
  Future<Response> apiCall(NetworkState state) async {
    Uri finalEndpoint = endpoint(state.https, getFinalEndpoint(url, state));
    print("PUT $finalEndpoint");
    // print(body);
    return await put(
      finalEndpoint,
      body: jsonEncode(body),
      headers: headers(state),
      encoding: Encoding.getByName("application/json"),
    );
  }

  PutAction(String path, this.body, {this.url = ""}) : super(path, "PUT");
}

mixin Anonymous {
  Map<String, String> headers(NetworkState _) => {};
}
